@@ -3,8 +3,10 @@ import importlib
 import json
 import logging
 import random
+import requests.exceptions
 import string
 import sys
+import time
 import traceback
 import uuid
 
@@ -97,9 +99,18 @@ class Account(models.Model):
         blank=True,
         on_delete=models.SET_NULL,
     )
+    balance = models.DecimalField(
+        max_digits=18,
+        decimal_places=9,
+        blank=True,
+        editable=False,
+        default=0,
+    )
+    name = models.CharField(max_length=100)
 
     def __str__(self):
-        return self.address
+        balance = int(self.balance) if self.balance else 0
+        return f'{self.name} {balance}tz'
 
     @property
     def provider(self):
@@ -109,85 +120,32 @@ class Account(models.Model):
     def private_key(self):
         return decrypt(self.crypted_key) if self.crypted_key else None
 
-    @property
-    def to_spool(self):
-        """Return transactions pending for action"""
-        return self.transactions_sent.exclude(
-            state__in=(
-                'held',
-                'done',
-                'deploy-aborted',
-                'watch-aborted',
-                'postdeploy-aborted',
-            ),
-        ).order_by('created_at').select_subclasses()
-
-    @property
-    def caller(self):
-        if '_djcall_caller' not in self.__dict__:
-            # get existing sender for this sender or create a new one
-            self._djcall_caller = Caller.objects.get_or_create(
-                callback='djtezos.models.sender_queue',
-                kwargs=dict(pk=str(self.pk)),
-            )[0]
-        return self._djcall_caller
-
-    def spool(self, force=False):
-        if not force and self.caller.running:
-            return
-        self.caller.spool('blockchain')
-
-    @property
-    def balance(self):
-        return self.get_balance()
-
     def get_balance(self):
         return self.provider.get_balance(self.address, self.private_key)
 
+    @property
+    def codename(self):
+        return self.blockchain.endpoint.rstrip('/').split('/')[-1]
+
+    def get_tzkt_api_url(self):
+        return f'https://api.{self.codename}.tzkt.io/v1/accounts/{self.address}'
+
     def get_tzkt_url(self):
-        code = self.blockchain.endpoint.rstrip('/').split('/')[-1]
-        return f'https://{code}.tzkt.io/{self.address}/'
+        return f'https://{self.codename}.tzkt.io/{self.address}/'
 
+    def generate_private_key(self):
+        if self.crypted_key or not self.owner:
+            return
 
-def sender_queue(pk):
-    acc = Account.objects.filter(pk=pk).first()
-    if not acc:
-        logger.error(f'{pk} does not exist anymore')
-        return
+        passphrase = ''.join(
+            random.choice(string.ascii_letters) for i in range(42)
+        )
 
-    tx = acc.to_spool.first()
+        self.address, private_key = (
+            self.blockchain.provider.create_wallet(passphrase)
+        )
 
-    if not tx:
-        logger.info(f'Account {pk} has no pending transaction')
-        return
-
-    if tx.state in ('deploy', 'deploying'):
-        tx.deploy_state()
-
-    elif tx.state in ('watch', 'watching'):
-        tx.watch_state()
-
-    elif tx.state in ('postdeploy', 'postdeploying'):
-        tx.postdeploy_state()
-
-    # restart myself until I shut down by myself because my work is done
-    acc.spool(force=True)
-
-
-def account_wallet(sender, instance, **kwargs):
-    if instance.crypted_key or not instance.owner:
-        return
-
-    passphrase = ''.join(
-        random.choice(string.ascii_letters) for i in range(42)
-    )
-
-    instance.address, private_key = (
-        instance.blockchain.provider.create_wallet(passphrase)
-    )
-
-    instance.crypted_key = encrypt(private_key)
-signals.pre_save.connect(account_wallet, sender=Account)
+        self.crypted_key = encrypt(private_key)
 
 
 class Blockchain(models.Model):
@@ -202,6 +160,8 @@ class Blockchain(models.Model):
     confirmation_blocks = models.IntegerField(default=0)
     description = models.TextField(blank=True)
     is_active = models.BooleanField(default=True)
+    max_level = models.PositiveIntegerField(default=None, blank=True, null=True)
+    min_level = models.PositiveIntegerField(default=None, blank=True, null=True)
 
     def __str__(self):
         return self.name
@@ -218,8 +178,10 @@ class Blockchain(models.Model):
 class TransactionQuerySet(InheritanceQuerySetMixin, models.QuerySet):
     def for_user(self, user):
         return self.filter(
-            Q(sender__owner=user) | Q(receiver__owner=user),
-        ).distinct()
+            Q(sender__owner=user)
+            | Q(receiver__owner=user)
+            | Q(users=user)
+        )
 
 
 class TransactionManager(InheritanceManagerMixin, models.Manager):
@@ -236,7 +198,6 @@ class Transaction(models.Model):
     sender = models.ForeignKey(
         'Account',
         related_name='transactions_sent',
-        blank=True,
         null=True,
         on_delete=models.CASCADE,
     )
@@ -247,6 +208,10 @@ class Transaction(models.Model):
         null=True,
         on_delete=models.CASCADE,
     )
+    users = models.ManyToManyField(
+        settings.AUTH_USER_MODEL,
+        blank=True,
+    )
     created_at = models.DateTimeField(
         null=True,
         blank=True,
@@ -283,24 +248,32 @@ class Transaction(models.Model):
         db_index=True,
     )
     args = models.JSONField(null=True, default=list, blank=True)
+    args_mich = models.JSONField(null=True, default=list, blank=True)
     amount = models.PositiveIntegerField(
         null=True,
         blank=True,
         help_text='Amount in xTZ',
         db_index=True,
     )
+    level = models.PositiveIntegerField(
+        null=True,
+        blank=True,
+        db_index=True,
+    )
+    last_fail = models.DateTimeField(
+        null=True,
+        blank=True,
+        auto_now_add=True,
+    )
 
     STATE_CHOICES = (
         ('held', _('Held')),
+        ('aborted', _('Aborted')),
         ('deploy', _('To deploy')),
         ('deploying', _('Deploying')),
-        ('deploy-aborted', _('Deploy aborted')),
+        ('retrying', _('Retrying')),
         ('watch', _('To watch')),
         ('watching', _('Watching')),
-        ('watch-aborted', _('Watch aborted')),
-        ('postdeploy', _('To post-deploy')),
-        ('postdeploying', _('Post-deploying')),
-        ('postdeploy-aborted', _('Postdeploy aborted')),
         ('done', _('Finished')),
     )
     state = models.CharField(
@@ -316,7 +289,16 @@ class Transaction(models.Model):
     objects = TransactionManager()
 
     def __str__(self):
-        return self.txhash or self.contract_name or str(self.pk)
+        if self.txhash:
+            return self.txhash
+        elif self.function:
+            return f'{self.contract_name}.{self.function}()'
+        elif self.contract_name:
+            return self.contract_name
+        elif self.amount:
+            return f'{self.amount}xTZ'
+        else:
+            return str(self.pk)
 
     @property
     def blockchain(self):
@@ -336,6 +318,7 @@ class Transaction(models.Model):
             not self.amount
             and not self.function
             and not self.contract_micheline
+            and not self.contract_address
         ):
             raise ValidationError('Requires amount, function or micheline')
 
@@ -348,10 +331,7 @@ class Transaction(models.Model):
         if self.state not in self.states:
             raise Exception('Invalid state', self.state)
 
-        result = super().save(*args, **kwargs)
-        if self.sender_id:
-            self.sender.spool()
-        return result
+        return super().save(*args, **kwargs)
 
     def call(self, **kwargs):
         return Transaction.objects.create(
@@ -359,69 +339,6 @@ class Transaction(models.Model):
             **kwargs
         )
 
-    def deploy_state(self):
-        self.state_set('deploying')
-        try:
-            self.txhash = self.deploy()
-        except Exception as exception:
-            self.error = str(exception)
-            if isinstance(exception, PermanentError):
-                logger.exception(f'{self} deploy permanent error {self.error}')
-                self.state_set('deploy-aborted')
-            else:
-                if isinstance(exception, TemporaryError):
-                    logger.info(f'{self} temporary error: {self.error}')
-                else:
-                    logger.exception(f'{self} deploy exception: {self.error}')
-                self.state_set('deploy')
-        else:
-            self.error = ''
-            self.state_set('watch')
-
-    def watch_state(self):
-        self.state_set('watching')
-        try:
-            self.watch()
-        except Exception as exception:
-            self.error = str(exception)
-            if isinstance(exception, PermanentError):
-                logger.exception(f'{self} watch permanent error {self.error}')
-                self.state_set('watch-aborted')
-            else:
-                if isinstance(exception, TemporaryError):
-                    logger.info(f'{self} temporary error: {self.error}')
-                else:
-                    logger.exception(f'{self} watch exception: {self.error}')
-                self.state_set('watch')
-        else:
-            self.error = ''
-            self.state_set('postdeploy')
-
-    def watch(self):
-        self.sender.blockchain.provider.watch(self)
-
-    def postdeploy_state(self):
-        self.state_set('postdeploying')
-        try:
-            self.postdeploy()
-        except Exception as exception:
-            self.error = str(exception)
-            if isinstance(exception, PermanentError):
-                logger.exception(f'{self} postdeploy permanent error {self.error}')
-                self.state_set('postdeploy-aborted')
-            else:
-                if isinstance(exception, TemporaryError):
-                    logger.info(f'{self} temporary error: {self.error}')
-                else:
-                    logger.exception(f'{self} postdeploy exception: {self.error}')
-                self.state_set('postdeploy')
-        else:
-            self.error = ''
-            self.state_set('done')
-
-    def postdeploy(self):
-        pass
-
     def state_set(self, state):
         self.state = state
         self.history.append([
@@ -435,13 +352,13 @@ class Transaction(models.Model):
         # django.db.connection.close()
         # close_old_connections()
 
-    def deploy(self):
-        return self.provider.deploy(self)
-
     def get_tzkt_url(self):
         code = self.blockchain.endpoint.rstrip('/').split('/')[-1]
         return f'https://{code}.tzkt.io/{self.txhash}/'
 
+    def get_better_url(self):
+        return f'https://better-call.dev/search?text={self.txhash or self.contract_address}'
+
 
 class ContractManager(TransactionManager):
     def get_queryset(self):
